"""
This file contains some utilities that are used in the tensiometer package.
"""

###############################################################################
# initial imports:

import numpy as np
import scipy
import scipy.special
from scipy.linalg import sqrtm
from getdist import MCSamples
import inspect

###############################################################################


def from_confidence_to_sigma(P):
    """
    Transforms a probability to effective number of sigmas.
    This matches the input probability with the number of standard deviations
    that an event with the same probability would have had in a Gaussian
    distribution as in Eq. (G1) of
    (`Raveri and Hu 18 <https://arxiv.org/pdf/1806.04649.pdf>`_).

    .. math::
        n_{\\sigma}^{\\rm eff}(P) \\equiv \\sqrt{2} {\\rm Erf}^{-1}(P)

    :param P: the input probability.
    :return: the effective number of standard deviations.
    :reference: https://arxiv.org/1806.04649
    """
    if (np.all(P < 0.) or np.all(P > 1.)):
        raise ValueError('Input probability has to be between zero and one.\n',
                         'Input value is ', P)
    return np.sqrt(2.)*scipy.special.erfinv(P)

###############################################################################


def from_sigma_to_confidence(nsigma):
    """
    Gives the probability of an event at a given number of standard deviations
    in a Gaussian distribution.

    :param nsigma: the input number of standard deviations.
    :return: the probability to exceed the number of standard deviations.
    :reference: https://arxiv.org/1806.04649
    """
    if (np.all(nsigma < 0.)):
        raise ValueError('Input nsigma has to be positive.\n',
                         'Input value is ', nsigma)
    return scipy.special.erf(nsigma/np.sqrt(2.))

###############################################################################


def from_chi2_to_sigma(val, dofs, exact_threshold=6):
    """
    Computes the effective number of standard deviations for a chi squared
    variable.
    This matches the probability computed from the chi squared variable
    to the number of standard deviations that an event with the same
    probability would have had in a Gaussian
    distribution as in Eq. (G1) of
    (`Raveri and Hu 18 <https://arxiv.org/pdf/1806.04649.pdf>`_).

    .. math::
        n_{\\sigma}^{\\rm eff}(x, {\\rm dofs}) \\equiv
        \\sqrt{2} {\\rm Erf}^{-1}({\\rm CDF}(\\chi^2_{\\rm dofs}(x)))

    For very high statistical significant events this function
    switches from the direct formula to an accurate asyntotic expansion.

    :param val: value of the chi2 variable
    :param dofs: number of degrees of freedom of the chi2 variable
    :param exact_threshold: (default 6) threshold of value/dofs to switch to
        the asyntotic formula.
    :return: the effective number of standard deviations.
    :reference: https://arxiv.org/1806.04649
    """
    # check:
    if (np.all(val < 0.)):
        raise ValueError('Input chi2 value has to be positive.\n',
                         'Input value is ', val)
    if (np.all(dofs < 0.)):
        raise ValueError('Input number of dofs has to be positive.\n',
                         'Input value is ', dofs)
    # prepare:
    x = val/dofs
    # if value over dofs is low use direct calculation:
    if x < exact_threshold:
        res = from_confidence_to_sigma(scipy.stats.chi2.cdf(val, dofs))
    # if value is high use first order asyntotic expansion:
    else:
        lgamma = 2*np.log(scipy.special.gamma(dofs/2.))
        res = np.sqrt(dofs*(x + np.log(2)) - (-4 + dofs)*np.log(x*dofs)
                      - 2*np.log(-2 + dofs + x*dofs) + lgamma
                      - np.log(2*np.pi*(dofs*(x + np.log(2)) - np.log(2*np.pi)
                               - (-4 + dofs)*np.log(x*dofs)
                               - 2*np.log(-2 + dofs + x*dofs) + lgamma)))
    #
    return res

###############################################################################


def KL_decomposition(matrix_a, matrix_b):
    """
    Computes the Karhunen-Loeve (KL) decomposition of the matrix A and B. \n
    Notice that B has to be real, symmetric and positive. \n
    The algorithm is taken from
    `this link <http://fourier.eng.hmc.edu/e161/lectures/algebra/node7.html>`_.
    The algorithm is NOT optimized for speed but for precision.

    :param matrix_a: the first matrix.
    :param matrix_b: the second matrix.
    :return: the KL eigenvalues and the KL eigenvectors.
    """
    # compute the eigenvalues of b, lambda_b:
    _lambda_b, _phi_b = np.linalg.eigh(matrix_b)
    # check that this is positive:
    if np.any(_lambda_b < 0.):
        raise ValueError('B is not positive definite\n',
                         'eigenvalues are ', _lambda_b)
    sqrt_lambda_b = np.zeros_like(_lambda_b)
    mask = _lambda_b > 0.0
    sqrt_lambda_b[mask] = 1.0 / np.sqrt(_lambda_b[mask])
    _sqrt_lambda_b = np.diag(sqrt_lambda_b)
    _phib_prime = np.dot(_phi_b, _sqrt_lambda_b)
    _a_prime = np.dot(np.dot(_phib_prime.T, matrix_a), _phib_prime)
    _lambda, _phi_a = np.linalg.eigh(_a_prime)
    _phi = np.dot(np.dot(_phi_b, _sqrt_lambda_b), _phi_a)
    return _lambda, _phi

###############################################################################


def QR_inverse(matrix):
    """
    Invert a matrix with the QR decomposition.
    This is much slower than standard inversion but has better accuracy
    for matrices with higher condition number.

    :param matrix: the input matrix.
    :return: the inverse of the matrix.
    """
    _Q, _R = np.linalg.qr(matrix)
    return np.dot(_Q, np.linalg.inv(_R.T))

###############################################################################


def clopper_pearson_binomial_trial(k, n, alpha=0.32):
    """
    http://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval
    alpha confidence intervals for a binomial distribution of k expected
    successes on n trials.

    :param k: number of success.
    :param n: total number of trials.
    :param alpha: (optional) confidence level.
    :return: lower and upper bound.
    """
    lo = scipy.stats.beta.ppf(alpha/2, k, n-k+1)
    hi = scipy.stats.beta.ppf(1 - alpha/2, k+1, n-k)
    return lo, hi

###############################################################################


def get_separate_mcsamples(chain):
    """
    Function that returns separate :class:`~getdist.mcsamples.MCSamples`
    for each sampler chain.

    :param chain: :class:`~getdist.mcsamples.MCSamples` the input chain.
    :return: list of :class:`~getdist.mcsamples.MCSamples` with the separate
        chains.
    """
    # get separate chains:
    _chains = chain.getSeparateChains()
    # copy the param names and ranges:
    _mc_samples = []
    for ch in _chains:
        temp = MCSamples()
        temp.paramNames = chain.getParamNames()
        temp.setSamples(ch.samples, weights=ch.weights, loglikes=ch.loglikes)
        temp.sampler = chain.sampler
        temp.ranges = chain.ranges
        temp.updateBaseStatistics()
        _mc_samples.append(temp.copy())
    #
    return _mc_samples

###############################################################################


def bernoulli_thin(chain, temperature=1, num_repeats=1):
    """
    Function that thins a chain with a Bernoulli process.

    :param chain: :class:`~getdist.mcsamples.MCSamples` the input chain.
    :param temperature: temperature of the Bernoulli process. If T=1 then
        this produces a unit weight chain.
    :param num_repeats: number of repetitions of the Bernoulli process.
    :return: a :class:`~getdist.mcsamples.MCSamples` chain with the
        reweighted chain.
    """
    # check input:
    # get the trial vector:
    test = np.log(chain.weights / np.sum(chain.weights))
    new_weights = np.exp((1. - temperature) * test)
    test = temperature*(test - np.amax(test))
    # do the trial:
    _filter = np.zeros(len(test)).astype(bool)
    _sample_repeat = np.zeros(len(test)).astype(int)
    for i in range(num_repeats):
        _temp = np.random.binomial(1, np.exp(test))
        _sample_repeat += _temp.astype(int)
        _filter = np.logical_or(_filter, _temp.astype(bool))
    new_weights = _sample_repeat*new_weights
    # filter the chain:
    chain.setSamples(samples=chain.samples[_filter, :],
                     weights=new_weights[_filter],
                     loglikes=chain.loglikes[_filter])
    # update:
    chain._weightsChanged()
    chain.updateBaseStatistics()
    #
    return chain

###############################################################################


def random_samples_reshuffle(chain):
    """
    Performs a coherent random reshuffle of the samples.

    :param chain: :class:`~getdist.mcsamples.MCSamples` the input chain.
    :return: a :class:`~getdist.mcsamples.MCSamples` chain with the
        reshuffled chain.
    """
    # check input:
    # get the reshuffling vector:
    _reshuffle_indexes = np.arange(len(chain.weights))
    np.random.shuffle(_reshuffle_indexes)
    # filter the chain:
    if hasattr(chain, 'loglikes') and chain.loglikes is not None:
        chain.setSamples(samples=chain.samples[_reshuffle_indexes, :],
                         weights=chain.weights[_reshuffle_indexes],
                         loglikes=chain.loglikes[_reshuffle_indexes])
    else:
        chain.setSamples(samples=chain.samples[_reshuffle_indexes, :],
                         weights=chain.weights[_reshuffle_indexes])
    # update:
    chain._weightsChanged()
    chain.updateBaseStatistics()
    #
    return chain

###############################################################################


def make_list(elements):
    """
    Checks if elements is a list.
    If yes returns elements without modifying it.
    If not creates and return a list with elements inside.

    :param elements: an element or a list of elements.
    :return: a list containing elements.
    """
    if isinstance(elements, (list, tuple)):
        return elements
    else:
        return [elements]

###############################################################################


def PDM_to_vector(pdm):
    """
    Transforms a positive definite matrix of dimension :math:`d \\times d`
    into an unconstrained vector of dimension :math:`d(d+1)/2`.
    This does not use the Cholesky decomposition since we need guarantee of
    strictly positive definiteness.

    The absolute values of the elements with indexes of the returned vector
    that satisfy:

    .. code-block:: python

        np.tril_indices(d, 0)[0] == np.tril_indices(d, 0)[1]

    are the eigenvalues of the matrix. The sign of these elements define
    the orientation of the eigenvectors.

    Note that this is not strictly the inverse of
    :meth:`tensiometer.utilities.stats_utilities.vector_to_PDM`
    since there are a number of discrete symmetries in the definition of the
    eigenvectors that we ignore since they are irrelevant for the sake of
    representing the matrix.

    :param pdm: the input positive definite matrix.
    :return: output vector representation.
    :reference: https://arxiv.org/abs/1906.00587
    """
    # get dimension:
    d = pdm.shape[0]
    # get the eigenvalues of the matrix:
    Lambda, Phi = np.linalg.eigh(pdm)
    # get triangular decomposition:
    (P, L, U) = scipy.linalg.lu(Phi.T, permute_l=False)
    # WR decomposition:
    Q, R = np.linalg.qr(L)
    L = np.dot(np.dot(R, U), L)
    # pivot the eigenvalues:
    Lambda2 = np.dot(np.dot(P.T, np.diag(Lambda)), P)
    # prepare output:
    mat = L
    mat[np.diag_indices(d)] = np.sign(mat[np.diag_indices(d)])*np.diag(Lambda2)
    #
    return mat[np.tril_indices(d, 0)]


def vector_to_PDM(vec):
    """
    Transforms an unconstrained vector of dimension :math:`d(d+1)/2`
    into a positive definite matrix of dimension :math:`d \\times d`.
    In the input vector the eigenvalues are in the positions where

    The absolute values of the elements with indexes of the input vector
    that satisfy:

    .. code-block:: python

        np.tril_indices(d, 0)[0] == np.tril_indices(d, 0)[1]

    are the eigenvalues of the matrix. The sign of these elements define
    the orientation of the eigenvectors.

    The purpose of this function is to allow optimization over the space
    of positive definite matrices that is either unconstrained or
    has constraints on the condition number of the matrix.

    :param pdm: the input vector.
    :return: output positive definite matrix.
    :reference: https://arxiv.org/abs/1906.00587
    """
    d = int(np.sqrt(1 + 8*len(vec)) - 1)//2
    L = np.zeros((d, d))
    # get the diagonal with eigenvalues:
    L[np.tril_indices(d, 0)] = vec
    Lambda2 = np.diag(np.abs(L[np.diag_indices(d)]))
    L[np.diag_indices(d)] = np.sign(L[np.diag_indices(d)]) * np.ones(d)
    # qr decompose L:
    Q, R = np.linalg.qr(L)
    Phi2 = np.dot(L, np.linalg.inv(R))
    # rebuild matrix
    return np.dot(np.dot(Phi2.T, Lambda2), Phi2)

###############################################################################


def whiten_samples(samples, weights):
    """
    Rescales samples by the square root of their inverse covariance.
    The resulting samples have identity covariance. This amounts to a change of
    coordinates so the physical meaning of different coordinates is changed.

    :param samples: the input samples.
    :param weights: the input weights of the samples.
    :return: whitened samples with identity covariance.
    """
    # compute sample covariance:
    _cov = np.cov(samples.T, aweights=weights)
    # compute its inverse square root:
    _temp = sqrtm(QR_inverse(_cov))
    # whiten the samples:
    white_samples = samples.dot(_temp)
    #
    return white_samples

###############################################################################


def is_outlier(points, thresh=3.5):
    """
    Returns a boolean array with True if points are outliers and False
    otherwise.

    :param points: An num-observations by num-dimensions array of observations
    :param thresh: The modified z-score to use as a threshold. Observations with
        a modified z-score (based on the median absolute deviation) greater
        than this value will be classified as outliers.
    :return: A num-observations-length boolean array.
    :reference: Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.
    """
    if len(points.shape) == 1:
        points = points[:, None]
    median = np.median(points, axis=0)
    diff = np.sum((points - median)**2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)
    if med_abs_deviation == 0:
        return np.zeros(diff.shape, dtype=bool)

    modified_z_score = 0.6745 * diff / med_abs_deviation

    return modified_z_score > thresh

###############################################################################


def filter_kwargs(dict_to_filter, function_with_kwargs):
    """
    Inspects a function signature and returns the correct kwargs. Usefull to
    isolate kwargs for a third party library.

    :param dict_to_filter: dictionary to filter
    :param function_with_kwargs: function to inspect
    """
    sig = inspect.signature(function_with_kwargs)
    filter_keys = [param.name for param in sig.parameters.values() if param.kind == param.POSITIONAL_OR_KEYWORD]
    filtered_dict = {filter_key: dict_to_filter[filter_key] for filter_key in filter_keys if filter_key in dict_to_filter.keys()}
    return filtered_dict



"""Tests for chain convergence diagnostics."""

#########################################################################################################
# Imports

import os
import unittest

import numpy as np
from getdist import MCSamples, loadMCSamples

import tensiometer.chains_convergence as conv
import tensiometer.utilities.stats_utilities as stut

#########################################################################################################


#########################################################################################################
# Convergence tests


class TestConvergence(unittest.TestCase):

    """Chain convergence diagnostics test suite."""
    def setUp(self):
        """Set up test fixtures."""
        self.here = os.path.dirname(os.path.abspath(__file__))
        self.chain = loadMCSamples(self.here + "/../../test_chains/DES")

    def test_gr_test_consistency(self):
        """Test Gelman-Rubin consistency for multiple chains."""
        res1 = conv.GR_test(self.chain)
        res2 = conv.GR_test(stut.get_separate_mcsamples(self.chain))
        assert np.allclose(res1[0], res2[0]) and np.allclose(res1[1], res2[1])
        res3 = conv.GR_test(
            stut.get_separate_mcsamples(self.chain),
            param_names=self.chain.getParamNames().getRunningNames(),
        )
        assert np.allclose(res1[0], res3[0]) and np.allclose(res1[1], res3[1])

    def test_gr_test_two_chains_consistency(self):
        """Test Gelman-Rubin consistency for two chains."""
        res2 = conv.GR_test(stut.get_separate_mcsamples(self.chain)[:2])
        res3 = conv.GR_test(
            stut.get_separate_mcsamples(self.chain)[:2],
            param_names=self.chain.getParamNames().getRunningNames(),
        )
        assert np.allclose(res2[0], res3[0]) and np.allclose(res2[1], res3[1])

    def test_grn_test_variants(self):
        """Test higher-moment Gelman-Rubin variants."""
        kwargs = {}
        print(conv.GRn_test(self.chain, n=2, param_names=None, feedback=2,
                            optimizer="ParticleSwarm", **kwargs))
        print(conv.GRn_test(self.chain, n=3, param_names=None, feedback=2,
                            optimizer="ParticleSwarm", **kwargs))
        print(conv.GRn_test(self.chain, n=2, param_names=None, feedback=2,
                            optimizer="TrustRegions", **kwargs))

    def test_grn_test_two_chains(self):
        """Test higher-moment Gelman-Rubin for two chains."""
        kwargs = {}
        print(conv.GRn_test(stut.get_separate_mcsamples(self.chain)[:2], n=2, param_names=None, feedback=0,
                            optimizer="ParticleSwarm", **kwargs))

    def test_validation_errors(self):
        """Test error handling for invalid inputs."""
        with self.assertRaises(TypeError):
            conv._helper_chains_to_chainlist(["not a chain"])
        with self.assertRaises(TypeError):
            conv._helper_chains_to_chainlist("not a chain container")
        with self.assertRaises(ValueError):
            conv._helper_chains_to_chainlist([stut.get_separate_mcsamples(self.chain)[0]])
        with self.assertRaises(ValueError):
            conv._helper_chains_to_chainlist([self.chain])
        with self.assertRaises(ValueError):
            conv.GRn_test_1D(self.chain, 2, param_name=['omegabh2', 'thetaMC'])
        with self.assertRaises(ValueError):
            conv.GRn_test_1D([self.chain], 2, param_name=['omegabh2'])

    def test_grn_theta0_branches(self):
        """Test theta0 branches in Gelman-Rubin n variants."""
        conv.GRn_test_1D(self.chain, 2, param_name="omegabh2", theta0=0.1)
        conv.GRn_test_1D_samples(
            [self.chain.samples[:, self.chain.index["omegabh2"]]] * 2,
            [self.chain.weights] * 2,
            3,
            theta0=0.0,
        )
        # n == 1 branch
        conv.GRn_test(self.chain, n=1)
        # single-parameter branch
        conv.GRn_test(self.chain, n=2, param_names=["omegabh2"])

    def test_gr_test_scalar_branch(self):
        """Test Gelman-Rubin scalar-parameter branch."""
        samples = [np.array([[0.0], [1.0]]), np.array([[0.5], [1.5]])]
        weights = [np.array([1.0, 1.0]), np.array([1.0, 1.0])]
        res, mode = conv.GR_test_from_samples(samples, weights)
        self.assertTrue(np.isfinite(res))
        self.assertEqual(mode.shape[0], 1)

    def test_grn_param_name_length_error(self):
        """Test parameter name length validation in GRn helpers."""
        samples = np.array([[0.0, 1.0], [1.0, 2.0]])
        names = ["x", "y"]
        labels = ["x", "y"]
        chain1 = MCSamples(samples=samples, names=names, labels=labels)
        chain2 = chain1.copy()
        with self.assertRaises(ValueError):
            conv.GRn_test_1D([chain1, chain2], 2, param_name=["x", "y"])

    def test_grn_from_samples_theta0_and_feedback(self):
        """Test GRn samples path with theta0 and feedback."""
        samples = [np.array([[0.0, 1.0]]), np.array([[1.0, 2.0]])]
        weights = [np.array([1.0]), np.array([1.0])]
        original = conv.teig.max_GtRq_brute

        def _stub_max_GtRq_brute(VM, MV, feedback=0, optimizer=None, **kwargs):
            """Stub brute optimizer output for GRn tests."""
            return ("stub", VM.shape, MV.shape)

        conv.teig.max_GtRq_brute = _stub_max_GtRq_brute
        try:
            result = conv.GRn_test_from_samples(
                samples, weights, n=2, theta0=np.array([0.5, 0.5]), feedback=1, optimizer="ParticleSwarm"
            )
            self.assertIsNotNone(result)
        finally:
            conv.teig.max_GtRq_brute = original

    def test_grn_from_samples_geap_branch(self):
        """Test GRn samples path with GEAP optimizer."""
        samples = [np.array([[0.0]]), np.array([[1.0]])]
        weights = [np.array([1.0]), np.array([1.0])]
        original = conv.teig.max_GtRq_geap_power

        def _stub_geap(VM, MV, **kwargs):
            """Stub GEAP optimizer output for GRn tests."""
            return ("geap", VM.shape, MV.shape)

        conv.teig.max_GtRq_geap_power = _stub_geap
        try:
            result = conv.GRn_test_from_samples(
                samples, weights, n=2, theta0=None, feedback=0, optimizer="GEAP"
            )
            self.assertIsNotNone(result)
        finally:
            conv.teig.max_GtRq_geap_power = original


#########################################################################################################
# Slow convergence tests


class TestChainsConvergenceSlow(unittest.TestCase):
    """Chain convergence slow test suite."""
    def setUp(self):
        """Set up test fixtures."""
        here = os.path.dirname(os.path.abspath(__file__))
        self.chain = loadMCSamples(here + "/../../test_chains/DES")

    def test_grn_test_geap(self):
        """Test GEAP optimizer for GRn convergence diagnostics."""
        subchains = stut.get_separate_mcsamples(self.chain)[:2]
        res = conv.GRn_test(subchains, n=2, param_names=None, feedback=0, optimizer="GEAP")
        self.assertIsNotNone(res)

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""Tests for flow_profiler utilities with lightweight stubs."""

#########################################################################################################
# Imports

import os
import types
import unittest
from unittest.mock import patch

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

#########################################################################################################
# Test configuration

os.environ["NUMBA_DISABLE_JIT"] = "1"

# Import after disabling JIT so numba honors the environment flag.
import tensiometer.synthetic_probability.flow_profiler as fp

tfb = tfp.bijectors

#########################################################################################################
# Helper flows


class TinyFlow:
    """Minimal flow stub for profiler."""

    def __init__(self):
        """Init."""
        self.num_params = 2
        self.param_names = ["p0", "p1"]
        self.param_labels = ["P0", "P1"]
        self.parameter_ranges = {"p0": (-1.0, 1.0), "p1": (-1.0, 1.0)}
        self.name_tag = "tiny"
        self.chain_samples = np.zeros((3, 2), dtype=np.float32)
        self.chain_loglikes = np.zeros(3, dtype=np.float32)

    def cast(self, arr):
        """Cast."""
        return tf.convert_to_tensor(arr, dtype=tf.float32)

    def sample(self, n):
        """Sample."""
        return tf.zeros((n, self.num_params), dtype=tf.float32)

    def log_probability(self, x):
        """Log probability."""
        return tf.zeros((tf.shape(x)[0],), dtype=tf.float32)

    def log_probability_jacobian(self, x):
        """Log probability jacobian."""
        return tf.zeros_like(x)

    def map_to_abstract_coord(self, x):
        """Map to abstract coord."""
        return x

    def map_to_original_coord(self, x):
        """Map to original coord."""
        return x


#########################################################################################################
# Unit tests


class TestFlowProfilerUnit(unittest.TestCase):
    """Flow profiler unit test suite."""
    def _profiler(self, **kwargs):
        """Build a profiler instance for unit tests."""
        return fp.posterior_profile_plotter(TinyFlow(), initialize_cache=False, feedback=0, **kwargs)

    def test_tiny_flow_methods(self):
        """Test TinyFlow helper methods."""
        flow = TinyFlow()
        arr = tf.zeros((2, flow.num_params), dtype=tf.float32)
        self.assertEqual(flow.log_probability_jacobian(arr).shape, arr.shape)
        self.assertTrue(np.allclose(flow.map_to_abstract_coord(arr).numpy(), arr.numpy()))
        self.assertTrue(np.allclose(flow.map_to_original_coord(arr).numpy(), arr.numpy()))

    def test_binned_argmax_utilities(self):
        """Test binned argmax utilities."""
        bins = np.array([0, 1, 0, 1])
        vals = np.array([0.1, 0.2, 0.3, 0.4])
        # call numba-wrapped functions via py_func to avoid JIT overhead in tests
        res1 = fp._binned_argmax_1D.py_func(bins, vals, 2)
        self.assertEqual(res1[0], 2)
        self.assertEqual(res1[1], 3)

        xb = np.array([0, 1])
        yb = np.array([1, 0])
        vals2 = np.array([0.5, 0.6])
        res2 = fp._binned_argmax_2D.py_func(xb, yb, vals2, 2, 2)
        self.assertEqual(res2[0, 1], 0)
        self.assertEqual(res2[1, 0], 1)

    def test_points_minimizer_tf_branch(self):
        """Test points_minimizer TensorFlow branch."""
        fake_res = types.SimpleNamespace(converged=True, objective_value=0.0, position=tf.zeros((2,)))
        with patch("tensiometer.synthetic_probability.flow_profiler.tfp.optimizer.lbfgs_minimize", return_value=fake_res):
            success, val, pt = fp.points_minimizer(lambda x: tf.constant(0.0), lambda x: x, np.array([0.0, 0.0]),
                                                   bounds=None, use_scipy=False)
        self.assertTrue(success)
        self.assertEqual(val, 0.0)
        self.assertTrue(np.allclose(pt, np.zeros(2)))

    def test_points_minimizer_scipy_branch(self):
        """Test points_minimizer SciPy branch."""
        fake_res = types.SimpleNamespace(success=True, fun=1.0, x=np.array([1.0, -1.0]), nfev=1, njev=1, message="ok")
        with patch("tensiometer.synthetic_probability.flow_profiler.minimize", return_value=fake_res) as mock_min:
            success, val, pt = fp.points_minimizer(lambda x: np.sum(x**2), lambda x: 2 * x,
                                                   [np.array([0.5, -0.5])], bounds=[(-1, 1), (-1, 1)],
                                                   use_scipy=True, feedback=2, use_jac=False)
        self.assertTrue(success[0])
        mock_min.assert_called()

    def test_par_and_number_missing(self):
        """Test _parAndNumber missing parameter handling."""
        prof = self._profiler()
        idx, name = prof._parAndNumber("missing")
        self.assertIsNone(idx)
        self.assertIsNone(name)

    def test_bestfit_and_likestats_errors(self):
        """Test best-fit and likestats error handling."""
        prof = self._profiler()
        with self.assertRaises(ValueError):
            prof.getBestFit()
        with self.assertRaises(ValueError):
            prof._initialize_likestats()

    def test_masked_gradient_ascent_runs(self):
        """Test masked gradient ascent runs."""
        prof = self._profiler()
        prof.temp_inv_cov = tf.eye(2, dtype=tf.float32)
        ensemble = tf.zeros((1, 2), dtype=tf.float32)
        mask = tf.constant([1.0, 0.0], dtype=tf.float32)
        out = prof._masked_gradient_ascent(learning_rate=0.1, num_iterations=0, ensemble=ensemble, mask=mask)
        self.assertEqual(out[0].shape[1], 2)

    def test_sample_profile_population_sets_temp(self):
        """Test sample profile population caching."""
        prof = self._profiler()
        with patch("tensiometer.synthetic_probability.flow_profiler.tf.linalg.inv",
                   return_value=tf.eye(2, dtype=tf.float32)):
            prof.sample_profile_population(num_minimization_samples=3)
        self.assertEqual(prof.temp_samples.shape[0], 3)
        self.assertEqual(prof.temp_probs.shape[0], 3)

    def test_find_flow_map_box_prior_path(self):
        """Test MAP search with a box prior."""
        prof = self._profiler(box_prior=True, use_scipy=False)
        prof.temp_samples = tf.zeros((3, 2))
        prof.temp_probs = tf.zeros((3,))
        prof.options["num_minimization_samples"] = 5
        fake_min = (np.array([True, True]), np.array([0.0, 0.0]), np.zeros((2, 2)))
        with patch.object(fp.posterior_profile_plotter, "_get_masked_box_bijector", return_value=tfb.Identity()), \
                patch("tensiometer.synthetic_probability.flow_profiler.points_minimizer", return_value=fake_min), \
                patch("tensiometer.synthetic_probability.flow_profiler.np.argpartition",
                      return_value=np.array([0, 1, 2])):
            val, sol = prof.find_MAP()
        self.assertEqual(sol.shape, (2,))

    def test_update_cache_no_updates(self):
        """Test update_cache no-op path."""
        prof = self._profiler()
        prof.update_cache(update_MAP=False, update_1D=False, update_2D=False)
        self.assertIsNone(prof.temp_samples)
        self.assertIsNone(prof.temp_probs)

    def test_update_cache_iterative_no_updates(self):
        """Test update_cache_iterative no-op path."""
        prof = self._profiler()
        prof.update_cache_iterative(update_1D=False, update_2D=False, niter=0)
        self.assertIsNone(prof.temp_samples)
        self.assertIsNone(prof.temp_probs)

#########################################################################################################
# Additional helper flows


class SmallFlow:
    """Small Flow test suite."""
    def __init__(self, with_ranges=True):
        """Init."""
        self.num_params = 2
        self.param_names = ["p0", "p1"]
        self.param_labels = ["P0", "P1"]
        self.parameter_ranges = {"p0": (0.0, 1.0), "p1": (0.0, 1.0)} if with_ranges else None
        self.name_tag = "sf"
        self.chain_samples = np.zeros((2, 2), dtype=np.float32)
        self.chain_loglikes = np.zeros(2, dtype=np.float32)

    def cast(self, arr):
        """Cast."""
        return tf.convert_to_tensor(arr, dtype=tf.float32)

    def sample(self, n):
        """Sample a simple grid."""
        return tf.constant(np.linspace(0, 1, n * self.num_params).reshape(n, self.num_params), dtype=tf.float32)

    def log_probability(self, x):
        """Log probability."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_abs(self, x):
        """Log probability abs."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_jacobian(self, x):
        """Log probability jacobian."""
        return tf.zeros((x.shape[0], self.num_params), dtype=tf.float32)

    def log_probability_abs_jacobian(self, x):
        """Log probability abs jacobian."""
        return tf.zeros((x.shape[0], self.num_params), dtype=tf.float32)

    def map_to_abstract_coord(self, x):
        """Map to abstract coord."""
        return x

    def map_to_original_coord(self, x):
        """Map to original coord."""
        return x


class TestFlowProfilerMore(unittest.TestCase):
    """Flow profiler additional test suite."""
    def test_small_flow_methods(self):
        """Test SmallFlow helper methods."""
        flow = SmallFlow(with_ranges=False)
        arr = tf.zeros((2, flow.num_params), dtype=tf.float32)
        self.assertEqual(flow.log_probability_abs(arr).shape, (2,))
        self.assertEqual(flow.log_probability_jacobian(arr).shape, (2, flow.num_params))
        self.assertEqual(flow.log_probability_abs_jacobian(arr).shape, (2, flow.num_params))
        self.assertTrue(np.allclose(flow.map_to_abstract_coord(arr).numpy(), arr.numpy()))
        self.assertTrue(np.allclose(flow.map_to_original_coord(arr).numpy(), arr.numpy()))

    def test_sample_profile_population_box_prior(self):
        """Test sample profile population with a box prior."""
        flow = SmallFlow(with_ranges=True)
        profiler = fp.posterior_profile_plotter(flow, initialize_cache=False, feedback=0, box_prior=True,
                                                use_scipy=False, pre_polish=False, polish=False, smoothing=False)
        with patch.object(profiler, "_get_masked_box_bijector", return_value=tfp.bijectors.Identity()), \
                patch("tensiometer.synthetic_probability.flow_profiler.tf.linalg.inv",
                      return_value=tf.eye(flow.num_params, dtype=tf.float32)):
            profiler.sample_profile_population(num_minimization_samples=4)
        self.assertIsNotNone(profiler.temp_cov)
        self.assertIsNotNone(profiler.temp_inv_cov)

    def test_find_map_scipy_path_randomized(self):
        """Test randomized MAP search with SciPy."""
        flow = SmallFlow(with_ranges=False)
        profiler = fp.posterior_profile_plotter(flow, initialize_cache=False, feedback=0, use_scipy=True,
                                                pre_polish=False, polish=False, smoothing=False)
        profiler.temp_samples = tf.constant([[0.0, 0.0], [0.5, 0.5]], dtype=tf.float32)
        profiler.temp_probs = tf.constant([0.1, 0.2], dtype=tf.float32)
        fake_pm = (np.array([True]), np.array([0.0]), np.array([[0.1, 0.2]]))
        with patch("tensiometer.synthetic_probability.flow_profiler.points_minimizer", return_value=fake_pm):
            val, sol = profiler.find_MAP(randomize=True, num_best_to_follow=1)
        self.assertEqual(sol.shape[0], 2)

    def test_masked_box_bijector_and_reset(self):
        """Test masked box bijector and cache reset."""
        flow = SmallFlow(with_ranges=True)
        profiler = fp.posterior_profile_plotter(flow, initialize_cache=False, feedback=0, use_scipy=False,
                                                pre_polish=False, polish=False, smoothing=False)
        bij = profiler._get_masked_box_bijector(mask=[1, 0])
        self.assertIsInstance(bij, tfp.bijectors.Bijector)
        profiler.profile_density_1D["p0"] = "filled"
        profiler.reset_cache()
        self.assertEqual(profiler.profile_density_1D, {})
        _ = fp._binned_argmax_1D.py_func(np.array([0, 1]), np.array([1.0, 2.0]), 3)
        _ = fp._binned_argmax_2D.py_func(np.array([0, 1]), np.array([0, 1]), np.array([1.0, 2.0]), 2, 2)

#########################################################################################################
# Additional helper flows


class DummyFlowAdditional:
    """Minimal flow stub to satisfy flow_profiler interfaces."""

    def __init__(self):
        """Init."""
        self.num_params = 2
        self.param_names = ["p0", "p1"]
        self.param_labels = ["P0", "P1"]
        self.parameter_ranges = {"p0": (-1.0, 1.0), "p1": (-2.0, 2.0)}
        self.name_tag = "dummy"
        self.chain_samples = np.zeros((2, 2), dtype=np.float32)
        self.chain_loglikes = np.zeros(2, dtype=np.float32)

    def cast(self, arr):
        """Cast."""
        return tf.convert_to_tensor(arr, dtype=tf.float32)

    def map_to_abstract_coord(self, x):
        """Map to abstract coord."""
        return x

    def map_to_original_coord(self, x):
        """Map to original coord."""
        return x

    def sample(self, n):
        """Sample."""
        return tf.zeros((n, self.num_params), dtype=tf.float32)

    def log_probability(self, x):
        """Log probability."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_abs(self, x):
        """Log probability abs."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_jacobian(self, x):
        """Log probability jacobian."""
        return tf.zeros((x.shape[0], self.num_params), dtype=tf.float32)

    def log_probability_abs_jacobian(self, x):
        """Log probability abs jacobian."""
        return tf.zeros((x.shape[0], self.num_params), dtype=tf.float32)


class TestFlowProfilerAdditional(unittest.TestCase):
    """Flow profiler targeted test suite."""
    def setUp(self):
        """Set up test fixtures."""
        # Stub TensorFlow minimizer globally to prevent long runs.
        fake_tf_res = types.SimpleNamespace(converged=True, objective_value=0.0, position=np.array([0.0, 0.0]))
        self._tf_patcher = patch("tensiometer.synthetic_probability.flow_profiler.tfp.optimizer.lbfgs_minimize",
                                 return_value=fake_tf_res)
        self._tf_patcher.start()

    def tearDown(self):
        """Clean up test fixtures."""
        self._tf_patcher.stop()

    def test_points_minimizer_scipy_and_tf_paths(self):
        """Exercise scipy and TensorFlow points_minimizer branches."""

        def func(x):
            """Simple quadratic objective."""
            return np.sum(x**2)

        def jac(x):
            """Gradient of the quadratic objective."""
            return 2 * x

        self.assertEqual(func(np.array([1.0, -1.0])), 2.0)
        self.assertTrue(np.allclose(jac(np.array([1.0, -1.0])), np.array([2.0, -2.0])))

        fake_res = types.SimpleNamespace(success=True, fun=1.0, x=np.array([0.0, 0.0]), nfev=1, njev=1, message="ok")
        with patch("tensiometer.synthetic_probability.flow_profiler.minimize", return_value=fake_res) as mock_min:
            success, val, pt = fp.points_minimizer(func, jac, [np.array([1.0, -1.0])],
                                                   bounds=[(-1, 1), (-1, 1)], feedback=2, use_scipy=True,
                                                   use_jac=False, method="L-BFGS-B",
                                                   options={"ftol": 1e-6, "gtol": 1e-5, "maxls": 10})
        self.assertTrue(success.all())
        mock_min.assert_called()
        success_tf, val_tf, pt_tf = fp.points_minimizer(func, jac, np.array([0.5, 0.5]),
                                                        bounds=[(-1, 1), (-1, 1)], use_scipy=False)
        self.assertTrue(success_tf)
        self.assertEqual(val_tf, 0.0)

    def test_dummy_flow_additional_methods(self):
        """Test DummyFlowAdditional helper methods."""
        flow = DummyFlowAdditional()
        arr = tf.zeros((2, flow.num_params), dtype=tf.float32)
        self.assertTrue(np.allclose(flow.map_to_abstract_coord(arr).numpy(), arr.numpy()))
        self.assertTrue(np.allclose(flow.map_to_original_coord(arr).numpy(), arr.numpy()))
        self.assertEqual(flow.sample(3).shape, (3, flow.num_params))
        self.assertEqual(flow.log_probability(arr).shape, (2,))
        self.assertEqual(flow.log_probability_abs(arr).shape, (2,))
        self.assertEqual(flow.log_probability_jacobian(arr).shape, (2, flow.num_params))
        self.assertEqual(flow.log_probability_abs_jacobian(arr).shape, (2, flow.num_params))

    def test_find_flow_MAP_branches(self):
        """Test find_flow_MAP branches."""
        flow = DummyFlowAdditional()
        with self.assertRaises(ValueError):
            fp.find_flow_MAP(flow, abstract=True, box_bijector=tfp.bijectors.Identity(), initial_points=tf.zeros((1, 2)))
        fake_res = (np.array([True, True]), np.array([0.0, -1.0]), np.array([[0.1, 0.2], [0.3, 0.4]]))
        with patch("tensiometer.synthetic_probability.flow_profiler.points_minimizer", return_value=fake_res):
            val, sol = fp.find_flow_MAP(flow, feedback=1, abstract=False, initial_points=tf.zeros((2, 2)),
                                        use_scipy=True)
        self.assertEqual(sol.shape[0], 2)
        with patch("tensiometer.synthetic_probability.flow_profiler.points_minimizer",
                   return_value=fake_res):
            val2, sol2 = fp.find_flow_MAP(flow, feedback=0, abstract=False,
                                          box_bijector=tfp.bijectors.Identity(),
                                          initial_points=tf.zeros((2, 2)),
                                          use_scipy=False)
        self.assertEqual(sol2.shape[0], 2)

    def test_posterior_profile_plotter_lightweight(self):
        """Test posterior_profile_plotter lightweight paths."""
        flow = DummyFlowAdditional()
        profiler = fp.posterior_profile_plotter(flow, initialize_cache=False, feedback=0, use_scipy=False,
                                                pre_polish=False, polish=False, smoothing=False)
        profiler.reset_cache()
        self.assertIsNone(profiler.temp_samples)
        profiler.update_cache(update_MAP=False, update_1D=False, update_2D=False)
        self.assertIsNone(profiler.temp_samples)

#########################################################################################################
# Extra helper flows


class DummyFlowExtra:
    """Minimal flow stub."""

    def __init__(self):
        """Init."""
        self.num_params = 2
        self.param_names = ["p0", "p1"]
        self.param_labels = ["P0", "P1"]
        self.parameter_ranges = {"p0": (-1.0, 1.0), "p1": (-1.0, 1.0)}
        self.name_tag = "dummy"
        self.chain_samples = np.zeros((3, 2), dtype=np.float32)
        self.chain_loglikes = np.zeros(3, dtype=np.float32)

    def cast(self, arr):
        """Cast."""
        return tf.convert_to_tensor(arr, dtype=tf.float32)

    def sample(self, n):
        """Sample."""
        return tf.zeros((n, self.num_params), dtype=tf.float32)

    def log_probability(self, x):
        """Log probability."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_jacobian(self, x):
        """Log probability jacobian."""
        return tf.zeros_like(x)

    def map_to_abstract_coord(self, x):
        """Map to abstract coord."""
        return x

    def map_to_original_coord(self, x):
        """Map to original coord."""
        return x


#########################################################################################################
# Extra tests


class TestFlowProfilerExtra(unittest.TestCase):
    """Flow profiler extra test suite."""
    def _make_profiler(self, **kwargs):
        """Build a profiler instance."""
        return fp.posterior_profile_plotter(DummyFlowExtra(), initialize_cache=False, feedback=0, **kwargs)

    def test_update_cache_cleanup_without_updates(self):
        """Test update_cache cleanup without updates."""
        profiler = self._make_profiler()
        profiler.update_cache(update_MAP=False, update_1D=False, update_2D=False)
        self.assertIsNone(profiler.temp_samples)
        self.assertIsNone(profiler.temp_probs)

    def test_reset_cache_clears_temp(self):
        """Test reset_cache clears cached state."""
        profiler = self._make_profiler()
        profiler.temp_samples = tf.ones((1, 2))
        profiler.temp_probs = tf.ones((1,))
        profiler.profile_density_1D = {"x": 1}
        profiler.reset_cache()
        self.assertIsNone(profiler.temp_samples)
        self.assertFalse(profiler.profile_density_1D)

    def test_dummy_flow_extra_methods(self):
        """Test DummyFlowExtra helper methods."""
        flow = DummyFlowExtra()
        arr = tf.zeros((2, flow.num_params), dtype=tf.float32)
        self.assertEqual(flow.cast(arr).shape, arr.shape)
        self.assertEqual(flow.sample(3).shape, (3, flow.num_params))
        self.assertEqual(flow.log_probability(arr).shape, (2,))
        self.assertEqual(flow.log_probability_jacobian(arr).shape, arr.shape)
        self.assertTrue(np.allclose(flow.map_to_abstract_coord(arr).numpy(), arr.numpy()))
        self.assertTrue(np.allclose(flow.map_to_original_coord(arr).numpy(), arr.numpy()))

    def test_stub_profiler_sampling(self):
        """Test StubProfiler sample_profile_population."""
        profiler = StubProfiler(DummyFlowIterative(with_ranges=True))
        profiler.sample_profile_population()
        self.assertEqual(profiler.temp_samples.shape, (2, 2))

    def test_update_cache_iterative_sets_bins(self):
        """Test update_cache_iterative sets bins."""
        profiler = self._make_profiler()
        profiler.index = {"p0": 0, "p1": 1}
        profiler._parAndNumber = lambda name: (0, types.SimpleNamespace(name="p0", label="p0"))

        def stub_sampler(self=None, **kwargs):
            """Stub sampler."""
            profiler.temp_samples = tf.zeros((2, 2), dtype=tf.float32)
            profiler.temp_probs = tf.zeros((2,), dtype=tf.float32)
            profiler.temp_inv_cov = tf.eye(2, dtype=tf.float32)

        stub_sampler()
        with patch.object(fp.posterior_profile_plotter, "sample_profile_population", new=stub_sampler):
            profiler.update_cache_iterative(params=["p0"], niter=0, update_1D=False, update_2D=False)
        self.assertIsNone(profiler.temp_samples)

    def test_initialize_likestats_profile_true(self):
        """Test _initialize_likestats with profile limits."""
        profiler = self._make_profiler()
        profiler.flow_MAP_logP = 0.0
        profiler.flow_MAP = np.zeros(2)
        profiler.temp_samples = tf.zeros((2, 2))
        profiler.temp_probs = tf.zeros((2,))
        density = types.SimpleNamespace(x=np.linspace(-1, 1, 3), P=np.ones(3))
        with patch.object(fp.mcsamples.MCSamples, "get1DDensityGridData", return_value=density):
            stats = profiler._initialize_likestats(profile_lims=True)
        self.assertIsNotNone(stats)
        self.assertTrue(hasattr(profiler, "likeStats"))

    def test_initialize_likestats_profile_false(self):
        """Test _initialize_likestats without profile limits."""
        profiler = self._make_profiler()
        profiler.flow_MAP_logP = 0.0
        profiler.flow_MAP = np.zeros(2)
        profiler.temp_samples = tf.zeros((3, 2))
        profiler.temp_probs = tf.zeros((3,))
        density = types.SimpleNamespace(x=np.linspace(-1, 1, 3), P=np.ones(3))
        with patch.object(fp.mcsamples.MCSamples, "get1DDensityGridData", return_value=density):
            stats = profiler._initialize_likestats(profile_lims=False)
        self.assertIsNotNone(stats)
        self.assertTrue(hasattr(stats, "ND_contours"))

    def test_getbestfit_raises_without_init(self):
        """Test getBestFit raises without initialization."""
        profiler = self._make_profiler()
        with self.assertRaises(ValueError):
            profiler.getBestFit()

    def test_getbestfit_after_initialize(self):
        """Test getBestFit after initialization."""
        profiler = self._make_profiler()
        profiler.flow_MAP_logP = 0.0
        profiler.flow_MAP = np.zeros(2)
        profiler._initialize_bestfit()
        bf = profiler.getBestFit()
        self.assertEqual(len(bf.names), profiler.n)

    def test_precompute_calls_get1d(self):
        """Test precompute_1D calls get1DDensityGridData."""
        profiler = self._make_profiler()
        with patch.object(fp.posterior_profile_plotter, "get1DDensityGridData") as mock_get1d:
            profiler.precompute_1D(["p0"])
        mock_get1d.assert_called_once()

    def test_get1d_density_else_branch(self):
        """Test get1DDensityGridData else branch."""
        profiler = self._make_profiler()
        profiler.index = {"p0": 0, "p1": 1}
        profiler._parAndNumber = lambda name: (0, types.SimpleNamespace(name="p0", label="p0"))
        cached_density = fp.Density1D(np.linspace(-1, 1, 3), np.ones(3))
        profiler.profile_density_1D = {0: cached_density}
        res = profiler.get1DDensityGridData("p0", num_points_1D=4)
        self.assertIsInstance(res, fp.Density1D)

    def test_permutations_iterable_branch_and_map_to_unitcube(self):
        """Test permutations iterable branch and map_to_unitcube."""
        prof = self._make_profiler(permutations=[np.array([0, 1])], map_to_unitcube=True, transformation_type="spline")
        # trigger the map_to_unitcube branch via cached bijector calls
        prof.temp_samples = tf.zeros((1, 2))
        prof.temp_probs = tf.zeros((1,))
        prof.flow_MAP_logP = 0.0
        prof.flow_MAP = np.zeros(2)
        prof._initialize_bestfit()
        self.assertIsNotNone(prof.bestfit)

#########################################################################################################
# Iterative helper flows


class DummyFlowIterative:
    """Minimal flow stub to avoid heavy computation."""

    def __init__(self, with_ranges=True):
        """Init."""
        self.num_params = 2
        self.param_names = ["p0", "p1"]
        self.param_labels = ["P0", "P1"]
        self.parameter_ranges = {"p0": (0.0, 1.0), "p1": (0.0, 1.0)} if with_ranges else None
        self.name_tag = "dummy"
        self.chain_samples = np.zeros((2, 2), dtype=np.float32)
        self.chain_loglikes = np.zeros(2, dtype=np.float32)

    def cast(self, arr):
        """Cast."""
        return tf.convert_to_tensor(arr, dtype=tf.float32)

    def sample(self, n):
        """Sample."""
        return tf.zeros((n, self.num_params), dtype=tf.float32)

    def log_probability(self, x):
        """Log probability."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_abs(self, x):
        """Log probability abs."""
        return tf.zeros((x.shape[0],), dtype=tf.float32)

    def log_probability_jacobian(self, x):
        """Log probability jacobian."""
        return tf.zeros((x.shape[0], self.num_params), dtype=tf.float32)

    def log_probability_abs_jacobian(self, x):
        """Log probability abs jacobian."""
        return tf.zeros((x.shape[0], self.num_params), dtype=tf.float32)

    def map_to_abstract_coord(self, x):
        """Map to abstract coord."""
        return x

    def map_to_original_coord(self, x):
        """Map to original coord."""
        return x


class StubProfiler(fp.posterior_profile_plotter):
    """Profiler subclass with stubbed sampling and densities."""

    def __init__(self, flow):
        """Init."""
        super().__init__(flow, initialize_cache=False, feedback=0, use_scipy=False,
                         pre_polish=False, polish=False, smoothing=False)

    def sample_profile_population(self, **kwargs):
        """Sample profile population."""
        self.temp_samples = tf.constant([[0.2, 0.3], [0.6, 0.7]], dtype=tf.float32)
        self.temp_probs = tf.constant([0.1, 0.2], dtype=tf.float32)
        self.temp_cov = tf.eye(self.n, dtype=tf.float32)
        self.temp_inv_cov = tf.eye(self.n, dtype=tf.float32)
        return None

    def get1DDensityGridData(self, name_or_idx, **kwargs):
        """Get1 D Density Grid Data."""
        return types.SimpleNamespace(x=np.array([0.0, 0.5, 1.0]))

    def get2DDensityGridData(self, ind1, ind2, **kwargs):
        """Get2 D Density Grid Data."""
        return types.SimpleNamespace(x=np.array([0.0, 0.5, 1.0]),
                                     y=np.array([0.0, 0.5, 1.0]))


#########################################################################################################
# Iterative tests


class TestFlowProfilerIterative(unittest.TestCase):
    """Flow profiler iterative test suite."""
    def setUp(self):
        """Set up test fixtures."""
        # Stub minimizers globally for safety.
        self._points_patch = patch(
            "tensiometer.synthetic_probability.flow_profiler.points_minimizer",
            return_value=(np.array([True]), np.array([0.0]), np.array([[0.0, 0.0]])),
        )
        self._points_patch.start()
        self._tf_patch = patch(
            "tensiometer.synthetic_probability.flow_profiler.tfp.optimizer.lbfgs_minimize",
            return_value=types.SimpleNamespace(converged=True, objective_value=0.0, position=np.array([0.0, 0.0])),
        )
        self._tf_patch.start()

    def tearDown(self):
        """Clean up test fixtures."""
        self._tf_patch.stop()
        self._points_patch.stop()

    def test_binned_argmax_helpers(self):
        """Test binned argmax helpers."""
        vals = np.array([1.0, 2.0, 3.0])
        bins = np.array([0, 1, 1])
        res1 = fp._binned_argmax_1D.py_func(bins, vals, 3)
        self.assertTrue(res1[1] >= 0)
        xb = np.array([0, 1, 0])
        yb = np.array([0, 1, 1])
        res2 = fp._binned_argmax_2D.py_func(xb, yb, vals, 2, 2)
        self.assertTrue(np.any(res2 >= 0))

    def test_update_cache_iterative(self):
        """Test update_cache_iterative."""
        profiler = StubProfiler(DummyFlowIterative())
        dummy1d = types.SimpleNamespace(x=np.array([0.0, 0.5, 1.0]))
        dummy2d = types.SimpleNamespace(x=np.array([0.0, 0.5, 1.0]), y=np.array([0.0, 0.5, 1.0]))
        with patch("getdist.mcsamples.MCSamples.get1DDensityGridData", return_value=dummy1d), \
                patch("getdist.mcsamples.MCSamples.get2DDensityGridData", return_value=dummy2d), \
                patch.object(fp.posterior_profile_plotter, "_parAndNumber", side_effect=lambda name: (0, name)):
            profiler.update_cache_iterative(params=["p0", "p1"], niter=0, update_1D=True, update_2D=True)
        self.assertIn("p0", profiler._1d_samples)
        self.assertIn(("p0", "p1"), profiler._2d_samples)
        profiler.reset_cache()
        self.assertEqual(profiler.profile_density_1D, {})

    def test_find_map_and_box_bijector(self):
        """Test find_MAP and box bijector branches."""
        flow = DummyFlowIterative(with_ranges=False)
        profiler = fp.posterior_profile_plotter(flow, initialize_cache=False, feedback=0, use_scipy=False,
                                                pre_polish=False, polish=False, smoothing=False)
        val, sol = profiler.find_MAP(x0=tf.zeros((1, flow.num_params)), randomize=False, abstract=True)
        self.assertEqual(sol.shape[0], flow.num_params)
        flow_with_ranges = DummyFlowIterative(with_ranges=True)
        profiler2 = fp.posterior_profile_plotter(flow_with_ranges, initialize_cache=False, feedback=0, use_scipy=False,
                                                 pre_polish=False, polish=False, smoothing=False, box_prior=True)
        bij = profiler2._get_masked_box_bijector(mask=[1, 0])
        self.assertIsInstance(bij, tfp.bijectors.Bijector)

    def test_dummy_flow_iterative_methods(self):
        """Test DummyFlowIterative helper methods."""
        flow = DummyFlowIterative(with_ranges=True)
        arr = tf.zeros((2, flow.num_params), dtype=tf.float32)
        self.assertEqual(flow.sample(2).shape, (2, flow.num_params))
        self.assertEqual(flow.log_probability(arr).shape, (2,))
        self.assertEqual(flow.log_probability_abs(arr).shape, (2,))
        self.assertEqual(flow.log_probability_jacobian(arr).shape, (2, flow.num_params))
        self.assertEqual(flow.log_probability_abs_jacobian(arr).shape, (2, flow.num_params))
        self.assertTrue(np.allclose(flow.map_to_abstract_coord(arr).numpy(), arr.numpy()))
        self.assertTrue(np.allclose(flow.map_to_original_coord(arr).numpy(), arr.numpy()))

#########################################################################################################
# Script entry point


if __name__ == "__main__":  # pragma: no cover
    unittest.main(verbosity=2)



"""
This file contains the loss functions for the normalizing flow training.

Since we are combining different loss functions we have different options.
"""

###############################################################################
# initial imports and set-up:

import numpy as np

# tensorflow imports:
import tensorflow as tf

###############################################################################
# helpers:

def _broadcast_sample_weight(sample_weight, losses):
    """
    Broadcast sample weights to match loss tensor shape.

    :param sample_weight: weights to apply.
    :param losses: loss tensor.
    :returns: broadcasted weights matching ``losses``.
    """
    if sample_weight is None:
        return None
    sample_weight = tf.cast(sample_weight, losses.dtype)
    if sample_weight.shape.rank == 0:
        return tf.ones_like(losses) * sample_weight
    if losses.shape.rank is None or sample_weight.shape.rank is None:
        losses_rank = tf.rank(losses)
        weight_rank = tf.rank(sample_weight)
        rank_diff = losses_rank - weight_rank
        new_shape = tf.concat([tf.shape(sample_weight), tf.ones(rank_diff, tf.int32)], axis=0)
        sample_weight = tf.reshape(sample_weight, new_shape)
        return tf.broadcast_to(sample_weight, tf.shape(losses))
    if sample_weight.shape.rank < losses.shape.rank:
        new_shape = sample_weight.shape.as_list() + [1] * (losses.shape.rank - sample_weight.shape.rank)
        sample_weight = tf.reshape(sample_weight, new_shape)
    return tf.broadcast_to(sample_weight, tf.shape(losses))


def _reduce_weighted_loss(losses, sample_weight, reduction):
    """
    Reduce weighted losses according to Keras reduction rules.

    :param losses: per-sample losses.
    :param sample_weight: sample weights for each loss.
    :param reduction: Keras reduction enum.
    :returns: reduced loss tensor.
    """
    weights = _broadcast_sample_weight(sample_weight, losses)
    if weights is not None:
        losses = losses * weights
    if reduction == tf.keras.losses.Reduction.NONE:
        return losses
    total_loss = tf.reduce_sum(losses)
    if reduction == tf.keras.losses.Reduction.SUM:
        return total_loss
    if weights is None:
        denom = tf.cast(tf.size(losses), losses.dtype)
    else:
        denom = tf.reduce_sum(weights)
    denom = tf.where(denom == 0, tf.constant(1.0, dtype=losses.dtype), denom)
    return total_loss / denom

###############################################################################
# standard normalizing flow loss function:


class standard_loss(tf.keras.losses.Loss):
    """KL-based density loss for normalizing flow training."""

    def __init__(self):
        """
        Standard density loss function for the normalizing flow.
        """
        # initialize:
        super(standard_loss, self).__init__()
        #
        return None

    def compute_loss_components(self, y_true, y_pred, sample_weight):
        """
        Compute the signed log-probability contribution.

        :param y_true: target log density (unused).
        :param y_pred: predicted log density.
        :param sample_weight: sample weights.
        :returns: negative predicted log density.
        """
        return -y_pred

    def call(self, y_true, y_pred):
        """
        Standard normalizing flow loss function is KL divergence of two abstract
        distributions.
        """
        return -y_pred
    
    def print_feedback(self, padding=''):
        """Print the configured loss details."""
        print(padding+'using standard loss function')

    def reset(self):
        """Reset loss hyperparameters (no-op for standard loss)."""
        pass


###############################################################################
# density and evidence loss with constant weights:


class constant_weight_loss(tf.keras.losses.Loss):
    """Combined density and evidence-error loss with fixed weights."""

    def __init__(self, alpha=1.0, beta=0.0):
        """
        Initialize the fixed-weight loss.

        :param alpha: weight for the density component.
        :param beta: additive offset applied to predicted log density.
        """
        # initialize:
        super(constant_weight_loss, self).__init__()
        # set parameters:
        self.alpha = alpha
        self.beta = beta
        #
        return None

    def compute_loss_components(self, y_true, y_pred, sample_weight):
        """
        Compute density and evidence-error components.

        :param y_true: target log density.
        :param y_pred: predicted log density.
        :param sample_weight: weights for each sample.
        :returns: tuple of density loss and variance of residuals.
        """
        # compute difference between true and predicted posterior values:
        diffs = (y_true - y_pred)
        # sum weights:
        tot_weights = tf.reduce_sum(sample_weight)
        # compute overall offset:
        mean_diff = tf.reduce_sum(diffs*sample_weight) / tot_weights
        # compute its variance:
        var_diff = (diffs - mean_diff)**2
        # compute density loss function:
        loss_orig = -(y_pred + self.beta)
        #
        return loss_orig, var_diff

    def compute_loss(self, y_true, y_pred, sample_weight):
        """
        Combine density and evidence-error loss.

        :param y_true: target log density.
        :param y_pred: predicted log density.
        :param sample_weight: weights for each sample.
        :returns: weighted sum of loss components.
        """
        # get components:
        loss_1, loss_2 = self.compute_loss_components(y_true, y_pred, sample_weight)
        #
        return +self.alpha*loss_1 + (1. - self.alpha)*loss_2

    def __call__(self, y_true, y_pred, sample_weight=None):
        """
        Override ``Loss.__call__`` to support explicit ``sample_weight``.

        :param y_true: target log density.
        :param y_pred: predicted log density.
        :param sample_weight: weights for each sample.
        :returns: weighted loss tensor respecting reduction mode.
        """
        if sample_weight is None:
            sample_weight = tf.ones_like(y_pred, dtype=y_pred.dtype)
        name_scope = getattr(self, "_name_scope", None)
        if not name_scope:
            name_scope = getattr(self, "name", self.__class__.__name__)
        with tf.name_scope(name_scope):
            losses = self.compute_loss(y_true, y_pred, sample_weight)
            return _reduce_weighted_loss(losses, sample_weight, self.reduction)

    def print_feedback(self, padding=''):
        """Print the configured fixed-weight loss settings."""
        print(padding+'using combined density and evidence-error loss function')
        print(padding+'weight of density loss: %.3g, weight of evidence-error-loss: %.3g' % (self.alpha, 1.-self.alpha))

    def reset(self):
        """Reset loss hyperparameters (no-op for constant weights)."""
        pass

###############################################################################
# density and evidence loss with variable weights:


class variable_weight_loss(tf.keras.losses.Loss):
    """Combined loss with trainable weights updated during training."""

    def __init__(self, lambda_1=1.0, lambda_2=0.0, beta=0.0):
        """
        Initialize the variable-weight loss.

        :param lambda_1: initial weight for the density term.
        :param lambda_2: initial weight for the evidence-error term.
        :param beta: additive offset applied to predicted log density.
        """
        # initialize:
        super(variable_weight_loss, self).__init__()
        # set parameters:
        self.lambda_1 = tf.Variable(lambda_1, trainable=False, name='loss_lambda_1', dtype=type(lambda_1))
        self.lambda_2 = tf.Variable(lambda_2, trainable=False, name='loss_lambda_2', dtype=type(lambda_2))
        self.beta = tf.Variable(beta, trainable=False, name='loss_beta', dtype=type(beta))
        # save initial parameters:
        self.initial_lambda_1 = lambda_1
        self.initial_lambda_2 = lambda_2
        self.initial_beta = beta
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...

        :param epoch: current epoch index.
        :param kwargs: unused passthrough arguments for compatibility.
        :raises NotImplementedError: expected to be overridden in subclasses.
        """
        # base class is empty...
        # use the following sintax:
        # tf.keras.backend.set_value(self.lambda_1, tf.constant(0.5) * epoch)
        raise NotImplementedError

    def compute_loss_components(self, y_true, y_pred, sample_weight, lambda_1=None, lambda_2=None):
        """
        Compute density and evidence-error components with configurable weights.

        :param y_true: target log density.
        :param y_pred: predicted log density.
        :param sample_weight: weights for each sample.
        :param lambda_1: optional override for density weight.
        :param lambda_2: optional override for evidence-error weight.
        :returns: tuple of density loss, variance of residuals, and active weights.
        """
        # compute difference between true and predicted posterior values:
        diffs = (y_true - y_pred)
        # sum weights:
        tot_weights = tf.reduce_sum(sample_weight)
        # compute overall offset:
        mean_diff = tf.reduce_sum(diffs*sample_weight) / tot_weights
        # compute its variance:
        var_diff = (diffs - mean_diff)**2
        # compute density loss function:
        loss_orig = -(y_pred + self.beta)
        # get weights if not passed:
        if lambda_1 is None:
            lambda_1 = tf.keras.backend.get_value(self.lambda_1)
        if lambda_2 is None:
            lambda_2 = tf.keras.backend.get_value(self.lambda_2)
        #
        return loss_orig, var_diff, lambda_1, lambda_2

    def compute_loss(self, y_true, y_pred, sample_weight):
        """
        Combine density and evidence-error loss using current weights.

        :param y_true: target log density.
        :param y_pred: predicted log density.
        :param sample_weight: weights for each sample.
        :returns: weighted loss value.
        """
        # get components:
        loss_1, loss_2, lambda_1,lambda_2 = self.compute_loss_components(y_true, y_pred, sample_weight, self.lambda_1, self.lambda_2)
        #
        return lambda_1*loss_1 + lambda_2*loss_2

    def __call__(self, y_true, y_pred, sample_weight=None):
        """
        Override ``Loss.__call__`` to support explicit ``sample_weight``.

        :param y_true: target log density.
        :param y_pred: predicted log density.
        :param sample_weight: weights for each sample.
        :returns: weighted loss tensor respecting reduction mode.
        """
        if sample_weight is None:
            sample_weight = tf.ones_like(y_pred, dtype=y_pred.dtype)
        name_scope = getattr(self, "_name_scope", None)
        if not name_scope:
            name_scope = getattr(self, "name", self.__class__.__name__)
        with tf.name_scope(name_scope):
            losses = self.compute_loss(y_true, y_pred, sample_weight)
            return _reduce_weighted_loss(losses, sample_weight, self.reduction)
        
    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        raise NotImplementedError

    def reset(self):
        """
        Reset loss functions hyper parameters
        """
        self.__init__(lambda_1=self.initial_lambda_1, lambda_2=self.initial_lambda_2, beta=self.initial_beta)


class random_weight_loss(variable_weight_loss):
    """
    Random weighting of the two loss functions.
    """

    def __init__(self, initial_random_epoch=0, lambda_1=1.0, beta=0.0, **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(random_weight_loss, self).__init__(lambda_1, 1.-lambda_1, beta)
        # set parameters:
        self.initial_random_epoch = initial_random_epoch
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        if epoch > self.initial_random_epoch:
            _temp_rand = np.random.randint(2)
            tf.keras.backend.set_value(self.lambda_1, _temp_rand)
            tf.keras.backend.set_value(self.lambda_2, 1.-_temp_rand)
        #        
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using randomized loss function')


class annealed_weight_loss(variable_weight_loss):
    """
    Slowly go from density to evidence-error loss.
    """

    def __init__(self, anneal_epoch=125, lambda_1=1.0, beta=0.0, roll_off_nepoch=10, **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(annealed_weight_loss, self).__init__(lambda_1, 1.-lambda_1, beta)
        # set parameters:
        self.anneal_epoch = anneal_epoch
        self.roll_off_nepoch = roll_off_nepoch
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        if epoch > self.anneal_epoch:
            _lambda_1 = tf.keras.backend.get_value(self.lambda_1)
            _lambda_1 *= np.exp(-1.*(epoch - self.anneal_epoch)/self.roll_off_nepoch)
            tf.keras.backend.set_value(self.lambda_1, _lambda_1)
            tf.keras.backend.set_value(self.lambda_2, 1.-_lambda_1)
        #
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using annealed loss function')


class SoftAdapt_weight_loss(variable_weight_loss):
    """
    Implement SoftAdapt as in arXiv:1912.12355, with optional smoothing
    """

    def __init__(self, tau=1.0, beta=0.0, smoothing=True, smoothing_tau=20, quantity_1='val_rho_loss', quantity_2='val_ee_loss', **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(SoftAdapt_weight_loss, self).__init__()
        # set parameters:
        self.tau = tau
        self.beta = beta
        self.smoothing = smoothing
        self.smoothing_alpha = 1. / smoothing_tau
        # quantrities to be monitored:
        self.quantity_1 = quantity_1 
        self.quantity_2 = quantity_2 
        self.rate_1_buffer = 0.0
        self.rate_2_buffer = 0.0
        self.lambda_1_buffer = 1.0
        self.lambda_2_buffer = 0.0
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        # get logs:
        logs = kwargs.get('logs')
        quantity_1 = logs[self.quantity_1]
        quantity_2 = logs[self.quantity_2]
        # get the two rates:
        if len(quantity_1) < 2:
            rate_1 = 0.0
        else:
            rate_1 = quantity_1[-1] - quantity_1[-2]
        if len(quantity_2) < 2:
            rate_2 = 0.0
        else:
            rate_2 = quantity_2[-1] - quantity_2[-2]
        # smooth the two rates:
        if self.smoothing:
            rate_1 = self.smoothing_alpha * rate_1 + (1. - self.smoothing_alpha) * self.rate_1_buffer
            rate_2 = self.smoothing_alpha * rate_2 + (1. - self.smoothing_alpha) * self.rate_2_buffer
            self.rate_1_buffer = rate_1
            self.rate_2_buffer = rate_2
        # protect for initial phase:
        if rate_1 == 0.0:
            _lambda_1 = 1.0
        else:
            lambda_1 = np.exp(self.tau * rate_1)
            lambda_2 = np.exp(self.tau * rate_2)
            _tot = lambda_1 + lambda_2
            if self.smoothing:
                _lambda_1 = self.smoothing_alpha * lambda_1 / _tot + (1. - self.smoothing_alpha) * self.lambda_1_buffer
                _lambda_2 = self.smoothing_alpha * lambda_2 / _tot + (1. - self.smoothing_alpha) * self.lambda_2_buffer
                _lambda_1 = _lambda_1 / (_lambda_1 + _lambda_2)
                _lambda_2 = _lambda_2 / (_lambda_1 + _lambda_2)
                self.lambda_1_buffer = _lambda_1
                self.lambda_2_buffer = _lambda_2
            else:
                _lambda_1 = lambda_1 / _tot
        # set second by enforcing sum to one:
        tf.keras.backend.set_value(self.lambda_1, _lambda_1)
        tf.keras.backend.set_value(self.lambda_2, 1.-_lambda_1)
        #
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using SoftAdapt loss function')
        if self.smoothing:
           print(padding+' with smoothing')
           print(padding+' with smoothing_tau = ', 1./self.smoothing_alpha)
        print(padding+' with tau = ', self.tau)


class SharpStep(variable_weight_loss):
    """
    Implement sharp stepping between two values
    """

    def __init__(self, step_epoch=50, value_1=1.0, value_2=0.1, beta=0., **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(SharpStep, self).__init__()
        # set parameters:
        self.step_epoch = step_epoch
        self.value_1 = value_1
        self.value_2 = value_2
        # initialize:
        self.beta = beta
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        if epoch < self.step_epoch:
            lambda_1 = self.value_1
        else:
            lambda_1 = self.value_2
        # set second by enforcing sum to one:
        lambda_2 = 1. - lambda_1
        #
        tf.keras.backend.set_value(self.lambda_1, lambda_1)
        tf.keras.backend.set_value(self.lambda_2, lambda_2)
        #
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using sharp step loss function')
